%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                   %
% Interpreter for focused intuitionistic linear logic (ILLF)        %
%                                                                   %
% Specific for AP                                                   %
%                                                                   %
% Authors: Elaine Pimentel and Carlos Olarte (2013)                 %
%                                                                   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% This prover is for a fragment of ILLF, used for specifying and
%% verifying properties of AP. The grammar we use is:

%% G :=  A | G ,G | sigma X\ G
%% P :=  pi X\ G -o P | P , P | sigma X\ P | P & P | bang pi X\ (p(X) -o  P) | one |  bang A | A.

%% The integer H in each rule is the limit of the height of the proof.
%% When doing focused proof search, one only need to use as the height of a proof 
% the number of decide rules encountered instead of all introduction rules.

% Sequents have the form AC, Gamma, Delta |- G
% where AC is the classical context containing only atoms, Gamma has only atomic and negative formulas,
% Delta is the linear context, and G is the goal.

% Observe that we are unfolding the definition clauses inside the body og the main formula.
% This will prevent unecessary calls to the theory and speed up the search.

% O is the output (for dealing with the tensor right and the implication left rules). It is inspired in Hodas and Miller's
% lazy passing of resources.


module illf.
accumulate blists.

%%%% Inference rules for ILLF

illf C Delta G :- prove_a C 0 nil nil Delta nil (G,top).

%Observe that this simulates guards (c , top), where c will be proven then top erases the junk.

%%% Negative Phase

prove_a C H AC Gamma (one :: Delta) O G :- 
        printf [nl, pc " One left", nl, fl],
        prove_a C H AC Gamma Delta O G.
            
prove_a C H AC Gamma ((A , B)::Delta) O G :- !,
        printf [nl, pc " Tensor left ", pc A, pc "' and '", pc B, pc "' put in the beginning of the list.", nl, fl],
        prove_a C H AC Gamma (A::B::Delta) O G.

prove_a C H AC Gamma ((sigma A)::Delta) O G :-
      printf [nl, pc " Exists left with expression ", pc A, pc "'", nl, fl], 
     pi X\ prove_pi C H AC Gamma ((A X)::nil) Delta O G.

prove_pi C H AC Gamma (B:: nil) Delta O G :- 
 printf [nl, pc " Delaying ", pc B, pc "'", nl, fl], 
prove_a C H AC Gamma (B::Delta) O G. 

% This is to move around a l-Prolog's BUG!!! 
% The correct would be prove_a H Gamma ((A X)::Delta) O G.
% But this does not work for more than one sigma in a row!!!

prove_a C H AC Gamma ((bang A)::Delta) O G :-
      printf [nl, pc " Bang left ", nl, fl], 
      prove_a C H (A::AC) Gamma Delta O G.

prove_a C H AC Gamma ((call DN LI)::Delta) O G :-
  DN LI FNew,
  prove_a C H AC  Gamma (FNew::Delta) O G.

      
%%% Synchronous Phase
            
prove_sr C H AC Gamma O (A , B):- 
        printf [nl, pc " Tensor right proving '", pc A, pc "' and '", pc B, pc "'", nl, fl],
    prove_sr C H AC Gamma M A, 
printf [nl, pc " Passing output context: ", pc M, nl, fl], 
    prove_sr C H AC M O B.
%printf [nl, pc O, pc " ", pc H, nl, fl].

prove_sr C H AC Gamma nil top:- 
        printf [nl, pc " 'Erasing contexts '", nl, fl].  
            
prove_sr C H AC Gamma O (sigma A) :-
        printf [nl, pc " Exists right with expression '", pc A, pc "'", nl, fl],
    sigma X\ prove_sr C H AC Gamma O (A X).

prove_sl C H AC (A -o  B) Gamma O G :- 
      printf [nl,  pc " Implication left '", pc A, pc "' -o '", pc B, pc "'",nl, fl], 
      prove_sr C H AC Gamma M A, 
      printf [nl, pc " Passing output context: ", pc M, nl, fl], 
      prove_sl C H AC B M O G.
%      printf [nl, pc O, pc " ", pc H, nl, fl]. 
      

prove_sl C H AC (pi A) Gamma O G :-
      printf [nl,  pc " For all left with expression '", pc A, pc "'", nl, fl], 
      sigma X\ prove_sl C H AC (A X) Gamma O G.
 

prove_sl C H AC (A & B) Gamma O G :-
      (printf [nl,  pc " Choosing on the left '", pc A, pc "'", nl, fl], 
      prove_sl C H AC A Gamma O G);
      (printf [nl,  pc " Choosing on the left '", pc B, pc "'", nl, fl], 
      prove_sl C H AC B Gamma O G).

%%% Reaction, Identity and Decide rules

%% Unfocusing on the left
prove_sl C H AC P Gamma O G :- 
       (atomic P,
        printf [nl, pc " Unfocusing on the left '", pc P, pc "' atomic", nl, fl], 
        NewGamma = (P::Gamma),
        print_g NewGamma, 
        prove_a C H AC NewGamma nil O G);
        (non_atomic P, pos P, 
        printf [nl, pc " Unfocusing on the left '", pc P, pc "' positive", nl, fl], 
        prove_a C H AC Gamma (P::nil) O G).

%% Collecting negative or atomic formulas on the left
prove_a C H AC Gamma (N::Delta) O G:- !,
    (neg N ; atomic N),
        printf [nl, pc" []L passing negative or atomic '", pc N, pc "' to Gamma.", nl, fl], 
    prove_a C H AC (N::Gamma) Delta O G.

%% DR Focusing on the right
prove_a C H AC Gamma nil O G :-  
     H1 is H + 1,
     H1 <= C,
        printf [nl, pc " DR rule: choosing '", pc G, pc "' positive or atomic on the right. H = ", pc H1, nl, fl], 
    prove_sr C H1 AC Gamma O G.


%% DL1 Focusing on the left linear context
prove_a C H AC Gamma nil O G :- 
    H1 is H + 1, 
    H1 <= C,
   memb_and_rest N Gamma NewGamma, 
   (non_atomic N), 
       printf [nl, pc " DL1, choosing '", pc N, pc "' negative from linear context. ", nl, fl], 
   prove_sl C H1  AC N NewGamma O G.


%% I

prove_sr C H AC Gamma O A :- 
    condition_init AC Gamma O A,
        printf [nl, pc " Initial rule: proving '", pc A, pc "' H = ",  pc H, nl, fl],
        print_g Gamma, print_o O,
        flush std_out.

condition_init AC Gamma O (ct X Num) :- 
	    memb_and_rest_num (ct X Num)  Gamma O.


condition_init AC Gamma O A :- 
            atomic A,
	    ((memb_and_rest A Gamma O);
            ((member A AC), equal Gamma O)).

              
%%%% Positive and negative formulas
pos (A , B).
pos (sigma A).
pos one.
pos (bang A).
pos (call X Y).
pos A :- atomic A.

pos top. % This is only for speeding up things. The LL connective top is negative, in general.

neg (A & B).
neg (pi A).
neg (A -o B).


%%%% Definition of atomic
non_atomic (A , B).
non_atomic (sigma A).
non_atomic one.
non_atomic (A & B).
non_atomic (pi A).
non_atomic top.
non_atomic (bang A).
non_atomic (A -o B).
non_atomic (call X Y).
atomic A :- not (non_atomic A).

%%%% Auxiliar functions

memb_and_rest_num (ct X Num)  ((ct X N) :: Rest) Rest :-
   Num is N.

memb_and_rest_num A (B :: Tail) (B :: Rest) :- memb_and_rest_num A Tail Rest.

% equal L M-> checks il L and M are the same list.
equal nil nil.
equal (X::L) (X::M) :- equal L M.


%%%% Prints

printf nil.
printf (pc (Item : string) :: L) :- !, print Item, printf L.
printf (pc Item :: L)            :- term_to_string Item Str, print Str, printf L.
printf (nl :: L)                 :- print "\n", printf L.
printf (fl :: L)                 :- flush std_out, printf L.

print_g G :- print "Context Gamma = {", term_to_string G StrG, print StrG, print "}\n".
print_o O :- print "Context O = {", term_to_string O StrG, print StrG, print "}\n".


%%%% Examples

pos a.
pos b.
pos d.
pos (c X).
pos (cc X).
pos (dd X Y).
atomic a.
atomic b.
atomic (c X).
atomic d.
atomic (cc X).
atomic (dd X Y).

%%%% TESTS
%equal (a::b::c::nil) X.
%filter (a::(a -o b)::b::(a,b)::nil) L M.
%illf nil (a::nil) a.
%illf nil ((a , b)::b::nil) (a , b).
%illf nil (a::(a -o b)::nil) b.
%illf nil (b::b::nil) b.
%illf nil ((a -o b)::nil) b.
%illf nil ((sigma c)::nil) (sigma c).
%illf nil ((a & b)::(bang d)::nil) b.
%illf nil ((bang d)::nil) d.
%illf 5 ((pi X\ ((c X) -o b))::nil) (((pi X\ (c X)),d)::nil) b.
%illf nil (((sigma c) -o (sigma c))::nil) a.
%illf nil ((sigma X \ (c X))::nil) (sigma Y\ (sigma X \ (c X))).
%illf nil ((sigma X \ (sigma Y\ (c X)))::nil) (sigma X \ (c X)).
%illf nil ((pi X \ (pi Y\ (c X)))::nil) (c a).
%body (pi X \ (pi Y\ ((c X) -o b))) X.
%illf 5 ((call defm1 (ng::nil))::(call defm2 (ng::ng::nil))::nil) (cc ng).
%illf 100 ((sigma X\ sigma Y\ ((c X), ((c Y) -o a)))::nil) a.
%illf 100 ((sigma X\ sigma Y\ ((cc Y), (c X), ((c X) -o a)))::nil) a.
%illf 100 ((pi X\ pi Y\ ((c X), ((c Y) -o a)))::nil) a.
%illf 100 ((sigma X\ (c X))::nil) (sigma Y\ (c Y)).

%illf 2 ((sigma X\ sigma Y\ ((cc Y), b, (c X), ((c X) -o a)))::nil) a.

%illf 5 ((sigma X\ sigma Y\ ((ct X (1 + 1)), b, (c X), ((c X) -o a)))::nil) (a, (ct X 2)).

%illf 100 ((sigma X\ sigma Y\ (b, (c X), ((c X) -o a)))::nil) a.

%illf 100 ((sigma X\ sigma Y\ (a, (c X), (a -o ((c X) -o b))))::nil) b.

%pi X\ (memb_and_rest a (a::X::nil) Y).
%pi X\ (memb_and_rest (c X) (a::(c X)::nil) Y).
%pi X\ (memb_and_rest a (a::(c X)::nil) Y).
%pi X\ (member a (X::a::nil)).
%pi X\ (delete_member a (X::a::nil) Y).
%pi X\ ((X::a::nil) = Y).
%pi X\ (equal (X::a::nil)  Y).

%pi X\ (memb_and_rest (c X) (a::(c X)::(body X X)::nil) Y).
%pi X\ (memb_and_rest (c X) (a::(c X)::nil) Y).
%pi X\ (memb_and_rest a (a::(c X)::nil) Y).
%pi X\ pi Z\ (memb_and_rest (c X) (a::(c X)::(c Z)::nil) Y).

